unit FlEdit;

(*********************************************************************************************************************

  TITLE:
   TFloatEdit is a descendant of TEdit to enter and check floating point numbers (min/max, number of decimal
   digits).
   TSpinFloatEdit is based on this, it also implements increment/decrement buttons and keystrokes and also
   "dynamic" inc/dec via an event handler.

  PACKAGE:
   TFloatEdit, TSpinFloatEdit components for borland delphi

  RELEASE:
   V 1.4, january 27, 2003

  TARGET:
   borland delphi 4/5/6/7?

  INSTALLATION:
   copy or extract fledit.pas and fledit.dcr into a folder included in delphi's search path
    (like $(DELPHI)\lib).
   select "component"->"install component" in delphi's main menu.
   type fledit.pas into the "unit name" edit box.
   select "OK".
   after the component package is opened in delphi, click "compile".
   then, after recompilation of the package, you should find the components on the "mirkes.de" component
    palette tab.

  USAGE:
   -TFloatEdit

     =added published properties:

       Digits: Byte (r/w)
        the number of decimal digits (digits after the decimal separator), range 0..18

       Min: Extended (r/w)
        the lowest valid value

       Max: Extended (r/w)
        the highest valid value

       Value: Extended (r/w)
        the current value

       ErrorMessage :string (r/w)
        the message that will be displayed if an invalid (not a number, out of range) value is typed in the edit
        box, if this is the empty string ("", default), no error message will be displayed, and the value will
        be reset to a default

       Alignment: TAlignment (r/w)
        how to align the text in the edit box (taLeftJustify, taCenter, taRightJustify (default))

       About: string (r)
        shows the version and copyright holder of the component in the delphi ide, not editable

     =added events:

       OnEnterPress : TKeyPressEvent (r/w)
        this event handler is called everytime the ENTER key is pressed

     =changed properties:

       Text: string (r)
        this property is not writeable in TFloatEdit, you can just read the current value in textual format. to
        change the value, use the Value property instead

   -TSpinFloatEdit
    TSpinFloatEdit inherits all properties of TCustomFloatEdit (like TFloatEdit), it additionally implements
    the following:

     =added types:

       TComputedIncrementEvent = procedure (Sender: TObject; const bIncrement:Boolean) of object;
        this event handler is used in the OnComputedIncrement (see below) event

     =added published properties:

       EditorEnabled: Boolean (rw)
        if this is set to flase (default is true), you cannot enter values in the edit box, you can just use the
        increment/decrement buttons and keys (up: CURSOR UP, down: CURSOR DOWN)

       Increment: Extended (rw)
        the value that is added to/substracted from the current value on increment/decrement event

     =added events:

       OnComputedIncrement : TComputedIncrement (r/w)
        if a method is assigned to this event handler, it will be called whenever an increment/decrement event
        happens rather than adding/substracting the Increment property's value. the boolean parameter bIncrement
        tells you whether an increment (true) or decrement (false) event is occured. in this method you may
        implement your own increment/decrement algorithm

  CONTACT:
   markus stephany, losheim am see, saarland, germany
   mailto:merkes@mirkes.de
   http://www.mirkes.de

  HISTORY:

   -Version 1.4, jan 27, 2003
    some problems with return/escape keys fixed, multiline style avoided on newer operating systems,
    on win98/me/2k/xp+ escape/return now are only handled by the component itself if no cancel/default
    buttons are on the form. thanks to:
    *Moz
    (this does not affect TSpinFloatEdit. it's always a multiline edit control)

   -Version 1.3c, nov 19, 2002
    entering numbers in scientific notation should now work better (reported by Scott Westfall and others)

   -Version 1.3b, oct 27, 2001
    fixed a bug: OnExit had been called twice, thanks to:
    *Philippe Oppliger
    included a db aware version of TFloatEdit 1.2 (TDBFloatEdit), written by:
    *Eddie Shipman, many thanks!

   -Version 1.3a, feb 23, 2001
    changed the license from GNU LGPL to the Clarified Artistic License

   -Version 1.3, jan 28, 2001
    implemented a base class TCustomFloatEdit where TFloatEdit is derived from
    hopefully fixed a problem with different decimalseparators
    text property is no longer writable (use value instead)
    added a component TSplinFloatEdit (like TSpinEdit from Borland's sample vcl, much of the code is just
    copy&paste'd from the sources, not just fix increment/decrement, also computed - via event handler)
    *credits to George Paraskevaidis
    *credits to Borland for the TSpinEdit sample component

   -Version 1.2
    Added properties: Alignment, OnEnterPress
    *credits to Giuseppe Mosca and Andrea Carli

   -Version 1.1
    There was a bug in the range checking; I hope it is fixed now. In V1.0 the value has been checked at each
    onkeypress-event, and that was not very fine in some cases. Now the correctness of the value will be checked
    when the focus get lost. If the value isn't in the range of min-max, the focus can't get lost. And now, the
    user can press the ESC-Key in FloatEdit to restore the value that was set before he made changes.


  LICENSE:
       TFloatEdit, TSpinFloatEdit Copyright (C)1997-2003 markus stephany

       This package is free software; you can redistribute it and/or modify
       it under the terms of either:

       a) the GNU General Public License as published by the Free
       Software Foundation; either version 2, or (at your option) any
       later version, or

       b) the Clarified Artistic License as used by the authors of NCFTP
       version 3.0.2.

       This package is distributed in the hope that it will be useful,
       but WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
       (Lesser) General Public License for more details.

       You should have received a copy of the GNU General Public License
       along with this package in the file named "COPYING"; if not,
       write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
       Boston, MA 02111-1307, USA.

       You should also have received a copy of the Clarified Artistic
       License along with this package in the file named "ARTISTIC"; if not,
       you may find a copy at http://www.mirkes.de/downloads/ARTISTIC.txt

********************************************************************************************************************)

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls,spin;

type
  TCustomFloatEdit = class(TCustomEdit)
  private
    { Private-Deklarationen }
    FDigits : Byte;
    FMin,FMax    : Extended;
    FDec         : Char;
    FErrorText     : string;
    FOldVal      : Extended;
    FAlignment   : TAlignment;
    FOnEnterPress: TKeyPressEvent;
    FTempValue   : Extended;
    function GetAbout: string; virtual;
    procedure SetAbout(const Value: string);
    function GetText: string;
    procedure ShowError;
  protected
    { Protected-Deklarationen }
    procedure CreateParams(var Params:TCreateParams); override;
    procedure SetValue(AValue : Extended);
    procedure SetMin(AValue : Extended);
    procedure SetMax(AValue : Extended);
    procedure SetDigits(AValue : Byte);
    function GetValue : Extended;
    procedure KeyPress(var Key: Char); override;
    procedure DoExit;override;
    procedure DoEnter;override;
    procedure SetAlignment(AValue:TAlignment);
    property OnEnterPress : TKeyPressEvent read FOnEnterPress write FOnEnterPress;
    property Digits   : Byte read FDigits write SetDigits;
    property Min : Extended read FMin write SetMin;
    property Max : Extended read FMax write SetMax;
    property Value    : Extended read GetValue write SetValue;
    property ErrorMessage :string read FErrorText write FErrorText;
    property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
    property About: string read GetAbout write SetAbout stored false;
  public
    { Public-Deklarationen }
    property Text: string read GetText;
    procedure Loaded; override;
  published
    { Published-Deklarationen }
    constructor Create (AOwner : TComponent);override;
  end;

  TFloatEdit = class (TCustomFloatEdit)
  published
    property About;
    property Digits;
    property Value;
    property Min;
    property Max;
    property ErrorMessage;
    property Alignment;
    property OnEnterPress;
    property Anchors;
    property AutoSelect;
    property AutoSize;
    property Color;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

  // for computed increment
  TComputedIncrementEvent = procedure (Sender: TObject; const bIncrement:Boolean) of object;

  TSpinFloatEdit = class (TCustomFloatEdit)
  private
    FIncrement: Extended;
    FButton: TSpinButton;
    FEditorEnabled : Boolean;
    FComputedIncrement: TComputedIncrementEvent;
    function GetMinHeight: Integer;
    procedure SetEditRect;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMPaste(var Message: TWMPaste);   message WM_PASTE;
    procedure WMCut(var Message: TWMCut);   message WM_CUT;
  protected
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
    procedure UpClick (Sender: TObject); virtual;
    procedure DownClick (Sender: TObject); virtual;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CheckIncDec(const bIncrement:Boolean);
    procedure CreateWnd; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Button: TSpinButton read FButton;
  published
    property About;
    property EditorEnabled: Boolean read FEditorEnabled write FEditorEnabled default True;
    property Increment: Extended read FIncrement write FIncrement;
    property OnComputedIncrement: TComputedIncrementEvent read FComputedIncrement write FComputedIncrement;
    property Digits;
    property Value;
    property Min;
    property Max;
    property ErrorMessage;
    property Alignment;
    property OnEnterPress;
    property Anchors;
    property AutoSelect;
    property AutoSize;
    property Color;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

procedure Register;

implementation

var
  // if the operating system is able to right align text in single line edits (win98 and above)
  // don't use the ES_MULTILINE style to avoid the escape key not triggering a cancel button
  SINGLE_LINE_CAN_RIGHT_ALIGN: Boolean;

// determine whether right alignment of text in single line edits is supported
// os >= win98 (win32,4,10) or >= 2000 (nt,5)
// called from initialization code
procedure CheckOS;
var
  recOS: TOSVersionInfo;
begin
  SINGLE_LINE_CAN_RIGHT_ALIGN := False;

  recOS.dwOSVersionInfoSize := sizeof(recOS);
  if GetVersionEx(recOS)
  then
    with recOS do
      case dwPlatformId of
        VER_PLATFORM_WIN32_WINDOWS:
          SINGLE_LINE_CAN_RIGHT_ALIGN := (dwMajorVersion > 4) or
            ((dwMajorVersion = 4) and (dwMinorVersion >= 10));
        VER_PLATFORM_WIN32_NT:
          SINGLE_LINE_CAN_RIGHT_ALIGN := (dwMajorVersion > 4);
      end;
end;

const
  NO_TEXT = '';
  SR_VER  = '1.4';
  SR_COPY = '1997-2003 markus stephany, merkes@mirkes.de, http://www.mirkes.de';

constructor TCustomFloatEdit.Create (AOwner : TComponent);
begin
  inherited Create(AOwner);
  FAlignment := taRightJustify;
  FDec := DecimalSeparator;
  FDigits := 1;
  FMin := 0;
  FMax := 100;//1E110;
  FErrorText := NO_TEXT;
  inherited Text := FloatToStrF(FMin,ffFixed,18,fDigits);
end;

function TCustomFloatEdit.GetAbout: string;
begin
  Result := self.ClassName+' '+SR_VER+','+SR_COPY
end;

procedure TCustomFloatEdit.SetAbout(const Value: string);
begin
  // no change
end;

function TCustomFloatEdit.GetText: string;
begin
  Result := inherited Text;
end;

procedure TCustomFloatEdit.Loaded;
begin
  inherited Loaded;
  Value := FTempValue;
end;

procedure TCustomFloatEdit.SetAlignment(AValue:TAlignment);
begin
  if FAlignment<>AValue then
  begin
    FAlignment:=AValue;
    RecreateWnd;
  end;
end;

procedure TCustomFloatEdit.CreateParams(var Params:TCreateParams);
const
  Alignments: array[Talignment] of Word= (ES_LEFT, ES_RIGHT, ES_CENTER);
begin
  inherited CreateParams(Params);
  Params.Style:=Params.Style or Alignments[FAlignment];

  // 1.4
  if not SINGLE_LINE_CAN_RIGHT_ALIGN
  then
    Params.Style := Params.Style or ES_MULTILINE;
end;

procedure TCustomFloatEdit.DoEnter;
begin
  if AutoSelect and not (csLButtonDown in ControlState) then
    SelectAll;
  FOldVal:=GetValue;
  inherited;
end;

procedure TCustomFloatEdit.DoExit;
var
  ts : string;
  Res: Extended;
begin
  ts := Text;
  if ts = ''
  then
    ts := FloatToStrF(FMin,ffFixed,18,fDigits);
  try
    Res := StrToFloat(ts);
    if (Res < FMin) or (Res > FMax)
    then
      Raise Exception.Create(FErrorText);
  except
    ShowError;
    Value := FOldVal;
    SelectAll;
    SetFocus;
    Exit;
  end;
  // Ausgabe formatieren
  inherited Text := FloatToStrF(Res,ffFixed,18,FDigits);
  Value:= StrToFloat(Text);
  inherited;
end;

procedure TCustomFloatEdit.SetValue(AValue : Extended);
var
  tmp : string;
begin
  if csLoading in ComponentState
  then
    FTempValue := AValue
  else
  begin
    if (AValue > FMax) then
    begin
      ShowError;
      AValue := FMax;
    end;
    if (AValue < FMin) then
    begin
      ShowError;
      AValue := FMin;
    end;
    tmp := FloatToStrF(AValue,ffFixed,18,FDigits);
    inherited Text:=tmp;
  end;
end;

function TCustomFloatEdit.GetValue : Extended;
var
  ts : string;
begin
  if csLoading in ComponentState
  then
    Result := FTempValue
  else
  begin
    ts := text;
    if (ts = '-') or (ts = FDec) or (ts = '') or (ts = 'E') then ts := '0';
    try
       Result := StrToFloat(ts);
    except
      Result := FMin;
    end;
    if Result < FMin
    then
       Result := FMin
    else
    if Result > FMax
    then
       Result := FMax
  end;
end;

procedure TCustomFloatEdit.SetDigits(AValue : Byte);
begin
  if FDigits <> AValue then
  begin
    if AValue > 18
    then
      AValue := 18;
    FDigits := AValue;
    Value := Value;
  end;
end;

procedure TCustomFloatEdit.SetMin(AValue : Extended);
begin
  if FMin <> AValue then
  begin
    if AValue > FMax
    then
      raise Exception.Create('Minimum value must not be greater than maximum value !');
    FMin := AValue;
    Value := Value;
  end;
end;

procedure TCustomFloatEdit.SetMax(AValue : Extended);
begin
  if FMax <> AValue then
  begin
    if AValue < FMin
    then
      raise Exception.Create('Maximum value must not be less than minimum value !');
    FMax := AValue;
    Value := Value;
  end;
end;


procedure TCustomFloatEdit.KeyPress(var Key: Char);
var
  ts : string;
  flTemp: Extended;
  intDigit: Integer;
begin
  if Key in [#10,#13] then
  begin
    if Assigned(FOnEnterPress) then FOnEnterPress(self,Key);
    //1.4 else // moz: remove else
      Key := #0;
    Exit;
  end;

  if Key = #27 then
  begin
    Value := fOldVal;
    SelectAll;
    //1.4 avoid 'bing'
    Key := #0;
    inherited;
    Exit;
  end;

  if Key < #32
  then
    inherited
  else
  begin
    if not (Key in ['0'..'9',FDec,'-','e']) then
    begin
      inherited;
      Key := #0;
      Exit;
   end;

   // prfen, ob neuer text im erlaubten bereich liegt
   // seltext durch key ersetzen
   ts := Copy(Text,1,SelStart)+Key+Copy(Text,SelStart+SelLength+1,MaxInt);
   if ts = '-' then
   begin
     inherited;
     Key := #0;
     inherited Text := '-0';
     SelStart := 1;
     SelLength:=1;
     Exit;
   end;

   if ts = FDec then
   begin
     inherited;
     Key := #0;
     inherited Text := '0'+FDec+'0';
     SelStart :=2;
     SelLength:=1;
     Exit;
   end;

   if ts = '' then
   begin
     inherited;
     Key := #0;
     inherited Text := FloatToStrF(FMin,ffFixed,18,FDigits);
     SelectAll;
     Exit;
   end;
   try
     flTemp := StrToFloat(ts);
   except
     inherited;
     Key := #0;
     Exit;
   end;

   if (flTemp < FMin) or
      (flTemp > FMax) or
      ((Key = FDec) and (FDigits = 0))
      then
   begin
     inherited;
     Key := #0;
     Exit;
   end;

   // anzahl der digits prfen (nicht bei exponentialschreibweise)
   if Pos('E', UpperCase(ts)) = 0 then
   begin
     intDigit := Pos(FDec, ts);
     if intDigit > 0
     then
      if Length(ts) - intDigit > FDigits then
      begin
        inherited;
        Key := #0;
        Exit;
      end;
   end;

   inherited;
   end;
end;

procedure TCustomFloatEdit.ShowError;
begin
  if FErrorText <> NO_TEXT
  then
    ShowMessage(FErrorText);
end;

{ TSpinFloatEdit }


constructor TSpinFloatEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FButton := TSpinButton.Create (Self);
  FButton.Width := 15;
  FButton.Height := 17;
  FButton.Visible := True;
  FButton.Parent := Self;
  FButton.FocusControl := Self;
  FButton.OnUpClick := UpClick;
  FButton.OnDownClick := DownClick;
  ControlStyle := ControlStyle - [csSetCaption];
  FIncrement := 1;
  FEditorEnabled := True;
end;

procedure TSpinFloatEdit.CheckIncDec(const bIncrement: Boolean);
begin
  if not Assigned(FComputedIncrement)
  then
    case bIncrement of
      False: Value := Value - FIncrement;
      True:  Value := Value + FIncrement;
    end
  else
    FComputedIncrement(Self, bIncrement);
end;

procedure TSpinFloatEdit.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  //1.4 EM_SETRECTNP only works on multiline edits
  Params.Style := Params.Style or WS_CLIPCHILDREN or ES_MULTILINE;
end;

procedure TSpinFloatEdit.CreateWnd;
begin
  inherited CreateWnd;
  SetEditRect;
end;

destructor TSpinFloatEdit.Destroy;
begin
  FButton := nil;
  inherited;
end;

procedure TSpinFloatEdit.DownClick(Sender: TObject);
begin
  if ReadOnly then MessageBeep(0)
  else
    CheckIncDec(False);
end;

procedure TSpinFloatEdit.GetChildren(Proc: TGetChildProc;
  Root: TComponent);
begin
end;

function TSpinFloatEdit.GetMinHeight: Integer;
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  DC := GetDC(0);
  GetTextMetrics(DC, SysMetrics);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  I := SysMetrics.tmHeight;
  if I > Metrics.tmHeight then I := Metrics.tmHeight;
  Result := Metrics.tmHeight + I div 4 + GetSystemMetrics(SM_CYBORDER) * 4 + 2;
end;

procedure TSpinFloatEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if not FEditorEnabled
  then
    Key := 0;
  if Key = VK_UP then UpClick (Self)
  else if Key = VK_DOWN then DownClick (Self);
  inherited KeyDown(Key, Shift);
end;

procedure TSpinFloatEdit.KeyPress(var Key: Char);
begin
  inherited;
  if not FEditorEnabled
  then
    Key := #0;
end;

procedure TSpinFloatEdit.SetEditRect;
var
  Loc: TRect;
begin
  SendMessage(Handle, EM_GETRECT, 0, LongInt(@Loc));
  Loc.Bottom := ClientHeight + 1;  {+1 is workaround for windows paint bug}
  Loc.Right := ClientWidth - FButton.Width - 2;
  Loc.Top := 0;
  Loc.Left := 0;
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@Loc));
  { debug SendMessage(Handle, EM_GETRECT, 0, LongInt(@Loc)); }
end;

procedure TSpinFloatEdit.UpClick(Sender: TObject);
begin
  if ReadOnly then MessageBeep(0)
  else
    CheckIncDec(True);
end;

procedure TSpinFloatEdit.WMCut(var Message: TWMCut);
begin
  if not FEditorEnabled or ReadOnly then Exit;
  inherited;
end;

procedure TSpinFloatEdit.WMPaste(var Message: TWMPaste);
begin
  if not FEditorEnabled or ReadOnly then Exit;
  inherited;
end;

procedure TSpinFloatEdit.WMSize(var Message: TWMSize);
var
  MinHeight: Integer;
begin
  inherited;
  MinHeight := GetMinHeight;
    { text edit bug: if size lesser than minheight, then edit ctrl does
      not display the text }
  if Height < MinHeight then
    Height := MinHeight
  else if FButton <> nil then
  begin
    if NewStyleControls and Ctl3D then
      FButton.SetBounds(Width - FButton.Width - 5, 0, FButton.Width, Height - 5)
    else FButton.SetBounds (Width - FButton.Width, 1, FButton.Width, Height - 3);
    SetEditRect;
  end;
end;

procedure Register;
begin
  RegisterComponents('mirkes.de', [TFloatEdit,TSpinFloatEdit]);
end;

initialization
  //1.4
  CheckOS;

end.